package org.jfree.chart.plot;

import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;

import org.jfree.chart.axis.NumberTick;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.axis.TickUnit;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.event.PlotChangeEvent;
import org.jfree.chart.renderer.xy.DefaultXYPolarItemRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.xy.XYDataset;
import org.jfree.ui.TextAnchor;

/**
 * Plots data that is in (theta, radius) pairs where theta equal to zero is due
 * east and increases anti-clockwise.
 */
public class XYPolarPlot extends XYPlot {
	/**
	 * Serialization version number
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * The tick unit that controls the spacing between the angular grid lines.
	 */
	private TickUnit angleTickUnit;

	/**
	 * Default constructor.
	 */
	public XYPolarPlot() {
		this(null, null, null, null);
	}

	/**
	 * Creates a new plot with the specified dataset, axes and renderer.
	 * 
	 * @param dataset
	 *            the dataset (<code>null</code> permitted).
	 * @param domainAxis
	 *            the domain axis (<code>null</code> permitted).
	 * @param rangeAxis
	 *            the range axis (<code>null</code> permitted).
	 * @param renderer
	 *            the renderer (<code>null</code> permitted).
	 */
	public XYPolarPlot(XYDataset dataset, ValueAxis domainAxis,
			ValueAxis rangeAxis, XYItemRenderer renderer) {
		super(dataset, domainAxis, rangeAxis, renderer);

		if (renderer != null) {
			this.setRenderer(0, renderer);
		}
		this.angleTickUnit = new NumberTickUnit(30.0D);
	}

	/**
	 * Returns the upper used bound of the radius axis.
	 * 
	 * @return The upper bound.
	 */
	public double getMaxUsedRadius() {
		double radius = 0.0;

		for (int nds = 0; nds < this.getDatasetCount(); nds++) {
			XYDataset ds = this.getDataset(nds);
			if (ds != null) {
				for (int nser = 0; nser < ds.getSeriesCount(); nser++) {
					for (int nit = 0; nit < ds.getItemCount(nser); nit++) {
						if (Math.hypot(ds.getXValue(nser, nit),
								ds.getYValue(nser, nit)) > radius) {
							radius = Math.hypot(ds.getXValue(nser, nit),
									ds.getYValue(nser, nit));
						}
					}
				}
			}
		}
		if (radius == 0.0) {
			radius = 1.0;
		}

		return radius;
	}

	/**
	 * Returns the upper bound of the radius axis.
	 * 
	 * @return The upper bound.
	 */
	public double getMaxRadius() {
		return Math.max(
				Math.max(Math.abs(this.getDomainAxis().getUpperBound()),
						Math.abs(this.getDomainAxis().getUpperBound())),
				Math.max(Math.abs(this.getRangeAxis().getUpperBound()),
						Math.abs(this.getRangeAxis().getUpperBound())));
	}

	/**
	 * Returns the tick unit that controls the spacing of the angular grid
	 * lines.
	 * 
	 * @return The tick unit (never <code>null</code>).
	 */
	public TickUnit getAngleTickUnit() {
		return this.angleTickUnit;
	}

	/**
	 * Sets the tick unit that controls the spacing of the angular grid lines,
	 * and sends a {@link PlotChangeEvent} to all registered listeners.
	 * 
	 * @param unit
	 *            the tick unit (<code>null</code> not permitted).
	 */
	public void setAngleTickUnit(TickUnit unit) {
		if (unit == null) {
			throw new IllegalArgumentException("Null 'unit' argument.");
		}
		this.angleTickUnit = unit;
		this.fireChangeEvent();
	}

	/**
	 * Generates a list of tick values for the angular tick marks.
	 * 
	 * @return A list of {@link NumberTick} instances.
	 */
	protected List refreshAngleTicks() {
		List ticks = new ArrayList();
		for (double currentTickVal = 0.0D; currentTickVal < 360.0D; currentTickVal += this.angleTickUnit
				.getSize()) {
			NumberTick tick = new NumberTick(new Double(currentTickVal),
					this.angleTickUnit.valueToString(currentTickVal),
					TextAnchor.CENTER, TextAnchor.CENTER, 0.0D);
			ticks.add(tick);
		}

		return ticks;
	}

	@Override
	protected void drawDomainGridlines(Graphics2D g2, Rectangle2D dataArea,
			List ticks) {
		if (this.getRenderer() == null) {
			return;
		}

		if (this.getRenderer() instanceof DefaultXYPolarItemRenderer) {
			((DefaultXYPolarItemRenderer) this.getRenderer())
					.drawAngularGridLines(g2, this, this.refreshAngleTicks(),
							dataArea);
		} else {
			super.drawDomainGridlines(g2, dataArea, ticks);
		}
	}

	@Override
	protected void drawRangeGridlines(Graphics2D g2, Rectangle2D dataArea,
			List ticks) {
		if (this.getRenderer() == null) {
			return;
		}

		if (this.getRenderer() instanceof DefaultXYPolarItemRenderer) {
			((DefaultXYPolarItemRenderer) this.getRenderer())
					.drawRadialGridLines(g2, this, this.getRangeAxis(), ticks,
							dataArea);
		} else {
			super.drawRangeGridlines(g2, dataArea, ticks);
		}
	}

	@Override
	public void setRenderer(int index, XYItemRenderer renderer, boolean notify) {
		super.setRenderer(index, renderer, notify);
		if (index == 0) {
			if (this.getRenderer() instanceof DefaultXYPolarItemRenderer) {
				this.getDomainAxis().setVisible(false);
				this.getRangeAxis().setVisible(false);
				this.setOutlineVisible(false);
			} else {
				this.getDomainAxis().setVisible(true);
				this.getRangeAxis().setVisible(true);
				this.setOutlineVisible(true);
			}

		}
	}
}